/** @file
    @brief  A resizeable data buffer used internally (implementation).
*/
#include "dt_buf.h"
#include <stdlib.h>
#include <string.h>

#ifndef NDEBUG
  /// The number of dt_buf instances that currently exist. A debugging facility.
  unsigned dt_buf_instance_count = 0;
#endif


dt_bool dt_buf_init( dt_buf* buf )
{
   enum{ initSize = 32 };
   dt_byte* begin = (dt_byte*)malloc(initSize);
   if( ! begin )
      return 0;
   buf->begin    = begin;
   buf->strEnd   = begin;
   buf->allocEnd = begin + initSize;
#ifndef NDEBUG
   ++ dt_buf_instance_count;
#endif
   return 1;
}

dt_bool  dt_buf_init_data( dt_buf* buf, dt_byte const* data, size_t len )
{
   dt_byte* begin = (dt_byte*)malloc(len);
   if( ! begin )
      return 0;
   memcpy( begin, data, len );
   buf->begin = begin;
   buf->strEnd = buf->allocEnd = begin + len;
#ifndef NDEBUG
   ++ dt_buf_instance_count;
#endif
   return 1;
}



void dt_buf_free( dt_buf* buf )
{
   free( buf->begin );
#ifndef NDEBUG
   -- dt_buf_instance_count;
#endif
}

void dt_buf_clear( dt_buf* buf )
{ buf->strEnd = buf->begin; }


dt_bool dt_buf_append( dt_buf* buf, int c )
{
   if( buf->strEnd == buf->allocEnd )
   {
      size_t oldSize = buf->strEnd - buf->begin;
      size_t newSize = (oldSize==0) ? 32 : (2*oldSize);
      dt_byte* newAddr = (dt_byte*)realloc( buf->begin, newSize );
      if( newAddr==0 )
         return 0;
      buf->begin    = newAddr;
      buf->strEnd   = newAddr+oldSize;
      buf->allocEnd = newAddr+newSize;
   }

   * buf->strEnd++ = c;
   return 1;
}


dt_bool  dt_buf_growth_reserve( dt_buf* buf, unsigned len )
{
   size_t avail = buf->allocEnd - buf->strEnd;
   if( len > avail )
   {
      dt_byte* newAddr;
      size_t oldSize = buf->strEnd - buf->begin;
      size_t newSize = (oldSize==0) ? 32 : (2*oldSize);
      if( newSize < oldSize+len ) newSize = oldSize+len;
      newAddr = (dt_byte*)realloc( buf->begin, newSize );
      if( newAddr==0 )
         return 0;
      buf->begin    = newAddr;
      buf->strEnd   = newAddr+oldSize;
      buf->allocEnd = newAddr+newSize;
   }
   return 1;
}


dt_bool dt_buf_append_data( dt_buf* buf, void const* data, unsigned len )
{
   if( ! dt_buf_growth_reserve( buf, len ) )
      return 0;
   memcpy( buf->strEnd, data, len );
   buf->strEnd += len;
   return 1;
}





