/** @file
    @brief  A resizeable data buffer used internally (interface).
*/
#ifndef INCLUDED_DT_BUF_H
#define INCLUDED_DT_BUF_H

#ifndef INCLUDED_DT_TYPE_H
#include "dt_type.h"
#endif


/** A structure that contains a dynamically allocated, resizeable block of memory. */
typedef struct dt_buf_
{
   dt_byte* begin;      /**< start of the allocated data (or NULL). */
   dt_byte* strEnd;     /**< end of the part of the memory block that is currently used. */
   dt_byte* allocEnd;   /**< end of the memory block that has been allocated. */
} dt_buf;

/**   Functions used to manipulate a data buffer contained in a dt_buf structure.
Function that return a dt_bool may fail to allocate the memory required for them to complete.
 - A non-zero value is returned if the operation was successful.
 - Zero is returned if a memory allocation failed. The buffer object is then left unmodified.
*/
/*@{*/
/** Initialize the buffer. @post begin == strEnd == a valid buffer address */
dt_bool  dt_buf_init( dt_buf* buf );
/** Initialize a buffer and store specified data into it. */
dt_bool  dt_buf_init_data( dt_buf* buf, dt_byte const* data, size_t len );
/** Frees the memory associated with the buffer. Must balance every call to dt_buf_init. */
void     dt_buf_free( dt_buf* buf );
/** Makes the buffer empty without releasing its memory. @post strEnd == begin */
void     dt_buf_clear( dt_buf* buf );
/** Appends a single byte of value c to the end of the buffer. */
dt_bool  dt_buf_append( dt_buf* buf, int c );
/** Ensure that the buffer allocates sufficient memory to store 'len' additional data.
@post  ( strEnd - begin )  has the same value as prior to the call
@post  ( allocEnd - strEnd )  >=  len
*/
dt_bool  dt_buf_growth_reserve( dt_buf* buf, size_t len );
/** Appends the data found at @a data of length @a len to the end of the buffer. */
dt_bool  dt_buf_append_data( dt_buf* buf, void const* data, size_t len );
/*@}*/

#endif/*INCLUDED_DT_BUF_H*/
