/** @file
    @brief  Public interface for in-memory handling of a DataTree structure.

@par Types of nodes in a datatree
In memory, a DataTree is handled as a tree of nodes,
which are accessible through a dt_node* pointer.
Each node can be either:
  - a leaf node containing a fixed-size chunk of data - created with dt_make_leaf()
  - an array of sub-nodes - created using dt_make_arr()
  - an record (or dictionary), where each sub-node is
    associated with a unique key - created using dt_make_rec()

@par Reference Counting
All nodes support reference-counting: multiple owners can share a node
by calling dt_node_addref(). <BR>
Newly created node have a reference count (or use count) initially set to 1. <BR>
Users of a node can release it by calling dt_node_release(). This will
decrement the use count, and delete the node if the use count reaches zero.

@note
      If you share the ownership of nodes in your code, be aware that
      no copy-on-write is automatically supported: if you modify a record
      or array using one reference, all other pointers that refer to it will be affected.

@par  Sub-Nodes
   When a sub-node is added to an array or record, the caller transfers
   node ownership to the parent. That is, the parent will release the
   node (by calling dt_node_release) when it is destroyed. <BR>
   This means that:
     - the caller who adds a node to a parent shall call dt_node_addref() if
       it wants to keep using the node after the parent has been released.
     - if the caller modifies the node after adding it to a parent,
       the node stored in the parent will itself be modified.
@note
   If a call that adds a sub-node to a parent fails and returns 0
   (for example because of a memory error), the child node is still released.
*/
#ifndef INCLUDED_DT_NODE_H
#define INCLUDED_DT_NODE_H

#include "dt_type.h"

/** A value that identifies the type of a node. */
typedef enum { dt_leaf_node, dt_arr_node, dt_rec_node } dt_node_kind;

typedef struct dt_node_ dt_node;

/** Node factory functions
Each of the following functions return a newly created node,
whose reference count is initially set to zero.
This means that the first call to dt_node_release will
release the memory associated with the node.
*/
/*@{*/
/** Returns a new leaf node containing the data starting at @a begin of length @a len. */
dt_node* dt_make_leaf( dt_byte const* begin, size_t len );
/** Returns a new empty record node. */
dt_node* dt_make_rec();
/** Returns a new empty array node. */
dt_node* dt_make_arr();
/*@}*/

/** Increment the use count of a node. */
void dt_node_addref ( dt_node* node );
/** Decrement the use count of a node, and free the node if the count reaches zero.
When the use count reaches zero, the memory allocated for the node is released,
and dt_node_release is called for all the direct sub-nodes it might contain.
*/
void dt_node_release( dt_node* node );

/**@TODO: Eventually add a call to make a node (recursively?) unique -- to support COW (copy-on-write). */

/** Returns the type of the node referred to by @a node. */
enum dt_node_kind dt_node_get_kind( dt_node* node );



/** Access the data associated with a leaf node.
@param   leaf    The leaf node whose contents are being queried.
@param   begin   The address of a pointer which, upon successful exit,
                  will point to the beginning of the data.
@param   len     The address of a variable which, upon successful exit,
                  will be set to the length of the node's data.
@returns A non-zero value if the call was successful. <BR>
         Zero will be returned if @a node is not a leaf node.
*/
dt_bool dt_leaf_get_data( dt_node const* leaf, dt_byte const** begin, size_t* len );



/** Returns the number of sub-nodes available in an array node.
A return value of zero indicates that the array is empty, or that @a arr is not an array node.
*/
size_t  dt_arr_get_size( dt_node const* arr );
/** Returns the sub-node of the array @a node located at index @a ind.
A NULL return value indicates that @a node is not an array node or that @a ind is out-of-range.
*/
dt_node*   dt_arr_get( dt_node const* arr, unsigned ind );
/** Assign a new sub-node to a specified existing index within the array.
@param arr   The array node in which an item is to be replaced
@param ind   The index, within the array, where the node is to be replaced.
@param item  The node to be inserted, for which the caller gives up ownership.
             ( this means it must be newly created, or dt_node_addref() must have been called ).
@return A non-zero value if @a node is an array and @a ind is a valid index. <BR>
        Zero if @a item could not be inserted, and has immediately been disposed of by calling dt_node_release().
*/
dt_bool           dt_arr_put( dt_node* arr, unsigned ind, dt_node* item );
/** Appends a new sub-node to the end of an array node.
@param arr   The array node in which an item is to be added.
@param item  The node to be inserted, for which the caller gives up ownership.
             ( this means it must be newly created, or dt_node_addref() must have been called ).
@return A non-zero value if @a node is an array and @a item was successfully inserted. <BR>
        Zero if memory could not be allocated, @a node is not an array, or @a item is null.
@note In case of failure, a non-null @a item will have been disposed of by calling dt_node_release().
*/
dt_bool  dt_arr_append( dt_node* arr, dt_node* item );



/** Returns the number of sub-nodes available in the record @a node.
A return value of zero indicates that the record is empty, or that @a node is not an record node.
*/
size_t  dt_rec_get_size( dt_node const* node );

/** Returns the sub-node of the record @a node which is associated with @a key (of length @a key_len).
A NULL return value indicates that @a node is not an array node,
or that the record does not contain a key matching (@a key,@a key_len).
*/
dt_node*   dt_rec_get( dt_node const* rec, dt_byte const* key, size_t key_len );

/** Identical to dt_rec_get(), but takes a zero-terminated string @a key to define the key. */
dt_node*   dt_rec_sget( dt_node const* rec, char const* key);


/** Put a sub-node associated with @a key (of length @a key_len) in the record @a node.
@param rec  The record in which the sub-node is to be placed.
@param key,key_len  The addess and legth of the key data with which the item is to be associated.
@param item  The sub-node to be added, for which the caller gives up ownership.
             ( this means it must be newly created, or dt_node_addref() must have been called ).
@return A non-zero value if @a node is a record and @a item was successfully inserted. <BR>
        Zero on failure, if memory could not be allocated or @a node is not a record --
          in which case @a item will have been disposed of by calling dt_node_release().
@note  If (@a key,@a key_len) was previously associated with another item,
       the previous item is replaced and released.
*/
dt_bool  dt_rec_put( dt_node* rec, dt_byte const* key, size_t key_len, dt_node* item );

/** Identical to dt_rec_get(), but takes a zero-terminated string @a key to define the key. */
dt_bool  dt_rec_sput( dt_node* rec, char const* key, dt_node* item );


/** Returns information about the @a ind -th item in the record @a node.
@return  The requested sub-node, or NULL if @a rec is not a record node or @a ind is out-of-range.
@param   rec   The record from which contents are queried
@param   ind   The index within @a rec of the record entry being queried.
@param   key   The optional address of a pointer which, if the return value is not NULL,
                will point to the beginning of the key associated with the node.
@param   key_len  The optional address of a size value which, if the return value is not NULL,
                will be set to the length of the key associated with the node.
*/
dt_node*   dt_rec_get_nth(  dt_node* rec, unsigned ind
                         , dt_byte const** key, size_t* key_len
                         );

#endif/*INCLUDED_DT_NODE_H*/

