#include "dt_stream_base.h"

#include <assert.h>


#ifndef NDEBUG
  /// The number of dt_reader instances that currently exist. A debugging facility.
  unsigned dt_reader_instance_count = 0;
  /// The number of dt_writer instances that currently exist. A debugging facility.
  unsigned dt_writer_instance_count = 0;
#endif


/** Retrieve a streaming step from @a obj into @a step.
This is a polymorphic function, which will use a subclass-specific
forwarding function stored within the object.
@return A non-zero value if the operation was successful.
*/
dt_bool dt_read ( dt_reader* obj, dt_stream_step* step )
{
   assert( (!!obj) && (obj->read_impl) && (obj->free_impl) );
   assert( !! step );
   
   return obj->read_impl( obj, step );
}

/** Destroy and free a dt_reader object. */
void dt_free_reader( dt_reader* obj )
{
   assert( (!!obj) && (obj->read_impl) && (obj->free_impl) );
   
   obj->free_impl( obj );
#ifndef NDEBUG
   -- dt_reader_instance_count;
#endif
}


/** Transmit the streaming step in @a step to @a obj.
This is a polymorphic function, which will use a subclass-specific
forwarding function stored within the object.
@return A non-zero value if the operation was successful.
*/
dt_bool dt_write( dt_writer* obj, dt_stream_step const* step)
{
   assert( (!!obj) && (obj->write_impl) && (obj->free_impl) );
   assert( !! step );
   
   return obj->write_impl( obj, step );
}

/** Destroy and free a dt_writer object. */
void dt_free_writer( dt_writer* obj )
{
   assert( (!!obj) && (obj->write_impl) && (obj->free_impl) );
   
   obj->free_impl( obj );
#ifndef NDEBUG
   -- dt_writer_instance_count;
#endif
}



void dt_init_reader_base( dt_reader* reader
                        , void (*free_impl)(dt_reader*)
                        , dt_bool (*read_impl)(dt_reader*, dt_stream_step*)
                        )
{
   reader->free_impl = free_impl;
   reader->read_impl = read_impl;
#ifndef NDEBUG
   ++ dt_reader_instance_count;
#endif
}

void dt_init_writer_base( dt_writer* writer
                        , void (*free_impl)(dt_writer*)
                        , dt_bool (*write_impl)(dt_writer*, dt_stream_step const*)
                        )
{
   writer->free_impl  = free_impl;
   writer->write_impl = write_impl;
#ifndef NDEBUG
   ++ dt_writer_instance_count;
#endif
}


