/** @file
    @brief  Tree streaming object that support a simple text file format
The objects that can be created using this header allow the conversion
of a data tree to/from a structured textual form.

The text format is defined as follows:
 - the beginning and end of a record are indicated using curly braces: { }
 - the beginning and end of an array are indicated using parenthesis: ( )
 - both record keys and leaf data are output using either of the following
   formats:
    - when only alphanumeric (and some specific special characters) are included,
      the text data is included as is. It is delimited by white space or
      any illegal identifier character. The characters allowed in such
      a naked identifier include:  A-Z a-z 0-9 . + - _
    - as a string quoted using the pascal language conventions:
        - one single-quote character starts and ends the string
        - single quote characters embedded in the string are doubled,
          which indicates that a single quote is part of the string.
   NB: more quoted formats may be included in the future
 - Record keys are immediately follows by a column character (:),
   which strictly identifies them, and is not allowed after a leaf value.
 - single-line comments can be included after a # character that is not
   embedded in a quoted strings.
 - For parsing purposes, blank characters (as indicated by the isspace() C function)
   and the comma (convenent to use a vector-like notation for arrays) are treated
   as whitespace, separating identifiers.

By default, the output text files are automatically indented for readability.
*/
#ifndef INCLUDED_DT_STREAM_TEXT_H
#define INCLUDED_DT_STREAM_TEXT_H

#include <stdio.h>

/** Formatting flags (currently only relevant to the output/writer object). */
typedef enum
   { dt_text_packed = 1<<0  /**< output in compact form (minimal spacing) */
   , dt_text_crlf   = 1<<1  /**< explicitly output CR+LF on every line end */
   } dt_text_mode_flags;

/** Create a reader that will decode the condents of the specified file.
The caller retains ownership of the source file handle (@a src),
and is responsible for closing it after destruction of the reader object.
@note This reader object only checks the structure, not the syntax, of the input file.
      For most purposes, it is useful to filter its ouput using
      a stream validator (@see dt_stream_validator).
*/
struct dt_reader_* dt_create_text_reader(FILE* src, dt_text_mode_flags mode);

/** Create a writer that will decode the condents of the specified file.
The caller retains ownership of the output file handle (@a src),
and is responsible for closing it after destruction of the writer object.
*/
struct dt_writer_* dt_create_text_writer(FILE* dst, dt_text_mode_flags mode);

#endif/*INCLUDED_DT_STREAM_TEXT_H*/

