#include "dt_stream_util.h"

#include "dt_stream_text.h"
#include "dt_stream_tree.h"
#include "dt_stream_validator.h"
#include "dt_node.h"

#include <stdio.h>


/** Reads stream steps from @a src and sends them to @a dst until completion or an error occurs.
@return Zero if either @a src or @a dst reported an error during the transmission.  <BR>
        A non-zero value if all steps including a final dt_end were transmitted.
@see dt_stream_validated_transfer() for a version of this function that can
        also be used if @a src might be sending an invalid data sequence
        (e.g. if the source is parsing an input file.
*/
dt_bool dt_stream_transfer( dt_reader* src, dt_writer* dst )
{
   dt_stream_step step;
   while(  dt_read ( src, &step )  &&  dt_write( dst, &step )  )
      if( step.kind == dt_end )
         return 1;
   return 0;
}


/** Reads stream steps from a reader to a writer while validating the structure of the stream.
The transfer is interrupted if the reader or writer report an error,
or if an inconsistency in the transmitted data structure is detected.
@return Zero if any error was encountered during the transmission.  <BR>
        A non-zero value if all steps including a final dt_end were transmitted.

This function is similar to dt_stream_transfer(), but automatically inserts a validator
to protect @a dst from structural errors in the stream received from @a src. <BR>
This is typically useful when the data source is a file being parsed.

@see dt_create_validator_writer -- the validator object which is used by this function.
*/
dt_bool dt_stream_validated_transfer( dt_reader* src, dt_writer* dst )
{
   dt_writer* validator = dt_create_validator_writer( dst );
   dt_bool ans =     dt_stream_transfer( src, validator )
                  && dt_validator_writer_completed(validator);
   dt_free_writer( validator );
   return ans;
}


/** Load the textual file at path @a srcPath into an in-memory data structure.
@return The root of the loaded datatree, or NULL if the operation failed.
*/
dt_node* dt_read_tree_from_file( const char* srcPath )
{
   dt_node* ans = 00;
   FILE* srcF = fopen(srcPath,"rb");
   if( srcF )
   {
      dt_reader* src = dt_create_text_reader( srcF, 0 );
      if( src )
      {
         dt_writer* treeMaker = dt_create_tree_writer( & ans );
         if( treeMaker )
         {
            if( ! dt_stream_validated_transfer( src, treeMaker ) )
            {
               dt_node_release( ans );
               ans = 0;
            }
            dt_free_writer(treeMaker);
         }
         dt_free_reader(src);
      }
      fclose(srcF);
   }
   return ans;
}


/** Saved the tree below @a root to a textual file at path @a srcPath.
@return True if the file was successfully written.
*/
dt_bool dt_write_tree_to_file( dt_node const* root, const char* dstPath )
{
#ifdef _WIN32
   enum { defaultFormat = dt_text_crlf };
#else
   enum { defaultFormat = 0 };
#endif
   dt_bool ans = 0;
   FILE* dstF = fopen(dstPath,"wb");
   if( dstF )
   {
      dt_writer* dst = dt_create_text_writer( dstF, defaultFormat );
      if( dst )
      {
         ans = dt_stream_tree( root, dst );
         dt_free_writer(dst);
      }
      ans = ( 0==fclose(dstF) )  &&  ans;
   }
   return ans;
}






